C QN.for: Contains procedures that implement the Quasi Newtion algorithm
C
C     QNP            Module that sets default values of the algorithm's parameters
C     CDGrad         Computes central difference gradient of y=f(n,x)
C     MinStep        Computes minimal allowed stepsize
C     GradTest       Computes the maximal relative gradient
C     QNStep         Performs line search
C     QuasiNewton1   Performs the Quasi Newton algorithm with update of the inverse hessian
C     BFGS           update of the inverse hessian
C     QuasiNewton2   Perfoms the Quasi Newton algorithm with update of the hessian
     

      Module QNP
      
	Logical QN_Print, QN_Global
	Real(8) QN_GradTol, QN_ParTol, QN_MaxIt

	Data QN_Print/.True./
	Data QN_Global/.True./
	Data QN_MaxIt/500./
	Data QN_GradTol/-1./
	Data QN_ParTol/-1./

	End Module QNP


C ***********************************************************************************
C
C CDGrad: Compute the central difference gradient of a user supplied function
C         y=f(n,x)
C
C Usage: Call CDGrad(n,x,f,df)
C
C Input: n, integer(4), the number of elements in x
C        x, real(8), n times 1 vector
C        f, pointer to the function y=f(n,x)
C
C Output: df, real(8) n times 1 vector
C
C Remarks: This is algorithm A.5.6.4. from Dennis and Schnabel (1983), p. 323
C          As there, I use eps**(1/3), instead of eps**(1/2) as in the manual of the IMSL library.
C          With this latter setting, the minization of the
C          Rosenbrock function - see QNTest.for - failed, at least when
C          I used the BFGS update of the hessian.

      Subroutine CDGrad(n,x,f,df)

	Integer(4) n
	Real(8) x(n), df(n)

      Integer(4) j
	Real(8) eps, h, temp, fx1, fx2, x1(n), x2(n), f
	External f

      x1=x      
      x2=x

	eps=Epsilon(eps)**(1.0/3.0)
C      eps=Epsilon(eps)**(1.0/2.0)
      
      do j=1,n
	   if (x1(j) .lt.0.0) then
	      h=-eps*dmax1(dabs(x1(j)),1.0)
	   else
	      h=eps*dmax1(dabs(x1(j)),1.0)
	   endif
	   temp=x1(j)
	   x1(j)=temp+h
	   x2(j)=temp-h
	   h=x1(j)-temp
	   fx1=F(n,x1)
	   fx2=F(n,x2)         
	   df(j)=(fx1-fx2)/(2.*h)	
	   x1(j)=x(j)
	   x2(j)=x(j)
	end do

	return

	End Subroutine

C***********************************************************************
C
C MinStep: Returns the minimal step size sl so that x1=x0+s*dx=0
C          (in terms of the parameter tolerance criterium)
C          
C Usage: sl=MinStep(nx,x0,dx)
C
C Input: nx, Integer(4), the number of elements in x0
C        x0, Real(8), nx times 1 vector
C        dx, Real(8), nx times 1 vector, change of x
C
C Output: sl, Real(8)

	Real(8) Function MinStep(nx,x0,dx) 

      use QNP

	Integer(4) nx
	Real(8) x0(nx), dx(nx)

	Integer(4) i
	Real(8) converge, temp

      converge=0.0

	Do i=1,nx
	   temp=dabs(dx(i))/dmax1(dabs(x0(i)),1.0)
	   if (temp .gt. converge) converge=temp
      end do

	MinStep=QN_ParTol/converge

	Return

	End Function

C***************************************************************************
C
C GradTest: implements the gradient criterium
C           from Dennis and Schnabel (1983), p. 160
C
C usage: converge=GradTest(nx,df,x,fx)
C
C Input:   nx, Integer(4), the number of elements in x0, x1
C          df, Real(8) vector with nx elements, the gradient of
C              f at x
C           x,  Real(8) vector with nx elements, the first value of x
C          fx, the value of f at x
C
C Output: Real(8), the maximal value of the convergence criterium

      Real(8) Function GradTest(nx,df,x,fx)
                        
	Integer(4) nx
	Real(8) df(nx), x(nx), fx

	Integer(4) i
	Real(8) converge, temp

      converge=0.0
	Do i=1,nx
	   temp=dabs(df(i))*dmax1(dabs(x(i)),1.0)
	   temp=temp/dmax1(dabs(fx),1.0)
         if (temp .gt. converge) converge=temp
      End Do

	GradTest=converge

	Return

	End Function

C QNStep
C
C Computes the step size so that the Quasi Newton algorithm
C always moves in the direction of a (local) minimum of f(x)
C
C usage: Call QNStep(N,x,dx,df,f0,smin,F,s)
C
C Input: N, Integer(4) the number of elements in x,
C        x, Real(8) N times 1 vector, the current parameter estimate
C       dx, Real(8) N times 1 vector, the current change of the parameter estimate
C       df, Real(8) N times 1 vector, the gradient of f(x) at x
C       f0, Real(8), equal to f(n,x), i.e., the value of f at the current x
C     smin, Real(8), the minimal step size
C       f2, Real(8), the value of f at x0+s*dx (the next iterate of x)
C        F, pointer the function y=f(n,x) 
C
C Output: s, Real(8) the step size found by the algorithm, if the algorithm fails, s=-1.0 is returned

                 
      Subroutine QNStep(N,x,dx,df,f0,sl,F,s,f2)

      use Errors
      use QNP

	Integer(4) N
	Real(8) x(N), dx(N), df(N), f0, sl, F, s, f2

      Real(8) smult, smin, smax, disc, s1, s2, amat(2,2), 
     +        bvec(1:2,1), ab(1:2,1), x1(N), f1, dfdx

      External F

C  Initialize 
      smult=1.0E-4
	smin=0.1E0
	smax=0.5E0
     
	dfdx=Dot_Product(df,dx)

      s1=1.0
	s=1.0
	FError=0		
      f1=f(n,x+dx)
	Do while ((FError .ne. 0) .and. (s1 .gt. sl)) ! backtracking until f can be evaluated at x+s1*dx
	   s1=0.75*s1
	   FError=0
	   f1=f(n,x+s1*dx)
	end do
	if (s1 .lt. sl) return
	dx=s1*dx
      s1=1.

C Try the full Newton step s=1 
      if (f1 .le. (f0 + smult*dfdx)) then
        s=s1
	 f2=f1
	 return
      else
        s=-dfdx/(2.0*(f1-f0-dfdx))
        if (s .lt. smin) s=smin
        if (s  .gt. smax) s=smax
        x1=x+s*dx
	  f2=f(n,x1)
      endif
      s2=s
	
C  Reduce s2 further unless g2 < g0 + s2*smult*dgdx 
      do while (f2 .gt. (f0 + smult*s2*dfdx) )

        amat(1,1)=1.0/(s2**2)
	  amat(1,2)=-1.0/(s1**2)
	  amat(2,1)=-s1/(s2**2)
	  amat(2,2)=s2/(s1**2)
        bvec(1,1)=f2-s2*dfdx-f0
	  bvec(2,1)=f1-s1*dfdx-f0
	  ab=MatMul(amat,bvec)
	  ab=ab/(s2-s1)
	  if (ab(1,1) .eq. 0.0) then
            s=-dfdx/(2*ab(2,1))
	  else
	      disc=(ab(2,1)**2)-3*ab(1,1)*dfdx	
	      if (disc .lt. 0.0) then
	         s=s2*smax
	       elseif (ab(2,1) .le. 0.0) then                 
	         s=(-ab(2,1)+dsqrt(disc))/(3*ab(1,1))
	       else
	         s=-dfdx/(ab(2,1)+dsqrt(disc))
	       endif
	  endif	          
      
        if (s .lt. s2*smin) s=s2*smin
        if (s .gt. s2*smax) s=s2*smax

        if (s .lt. sl) return
	    
      s1=s2
      s2=s

      f1=f2
      x1=x+s2*dx
      f2=f(n,x1)

      End Do

      Return

      End Subroutine
      
         

C*************************************************************************************************************
C QuasiNewton1
C
C Find the minimizer of a user supplied function y=f(nx,x)
C
C Usage: Call QuasiNewton(f,x0,x1,crit)
C
C Input:  f  : Real(8) pointer to f(nx,x) that returns the value of x, where nx is number of elements of x
C         x0 : Real(8) nx times 1 vector
C
C Output: x1 : Real(8) nx times 1 vector, the found solution
C       crit : Real(8) 5 times 1 vector, crit(1):=return code: 0: normal termination
C                                                              1: function evaluation at x not possible
C                                                              2: step2<Mstep, and no further reduction in f possible,
C                                        crit(2):=the maximum absolute value of the scaled gradient at x1                                     
C                                        crit(3):=the value of f at x1C                   
C                                        crit(4):=number of iterations
C
C  Gobals:     QN_Print    = .t. print to screen
C              QN_Global   = .t. use line search
C              QN_GradTol  = -1 use default value of MachEps**(1/3)
C              QN_ParTol   = -1 use default value of MachEps**(2/3)
C
C  Remarks: This program uses the updating formula for the inverse of the approximate Hessian matrix
C           given in Press et al. (1997), p. 420

      Subroutine QuasiNewton1(f,nx,x0,x1,crit)

      use DFLIB
	use Errors
	use QNP
	use DFLIB	
	use IMSLF90 

      Integer(4) nx
	Real(8) f,x0(nx),x1(nx),crit(4)

      Logical ishere
	Integer(2) fh
      Integer(4) i
	Real(8) step, mstep, f1, f2, df1(nx), df2(nx), dx(nx), ddf(nx)
	Real(8) MinStep, GradTest, hinv(nx,nx), x2(nx)
	 
	Type(rccoord) curpos

      External F
C Initialize
      Inquire(File='QNP.txt',Exist=ishere)
	if (ishere) then
	   Call GetD('qn_gradtol','QNP.txt',QN_GradTol)
	   Call GetD('qn_partol','QNP.txt',QN_ParTol)
	   Call GetD('qn_maxit','QNP.txt',QN_MaxIt)	   
	   Call GetL('qn_print','QNP.txt',QN_Print)
	   Call GetL('qn_global','QNP.txt',QN_Global)
      endif
C Set Default values, if not otherwise specified
      IF (QN_GradTol .lt. 0.) QN_GradTol=Epsilon(step)**(1./3.)
	IF (QN_ParTol  .lt. 0.) QN_ParTol=Epsilon(step)**(2./3.)

      x1=x0

C Initial function value and gradient
      FError=0
      f1=f(nx,x1)

	If (FError .ne. 0) then ! return,if x0 is not admissible
	   crit(1)=1
	   return
	endif

      Call CDGrad(nx,x1,f,df1)

C Test whether the current x0 is already a good solution  (as suggested by Dennis and Schnable (1983), p. 349))
      crit(2)=GradTest(nx,df1,x1,f1)
	if (crit(2) .lt. (1.0e-3)*QN_GradTol) then
	   crit(1)=0.0	
	   crit(3)=f1
	   return
	endif

C initialize inverse of approximate Hessian and compute Newton direction
      hinv=0.0
	Do i=1,nx
	   hinv(i,i)=1.0
	       dx(i)=-df1(i)
	end do

      mstep=MinStep(nx,x1,dx) 

      crit(4)=1.0

      if (QN_Print) then
         fh=GetActiveQQ() ! Handle of active window to write messages      
	   Call ClearScreen($GCLEARSCREEN)
	   CALL SETTEXTPOSITION(1,1,curpos)
	   WRITE(fh,*) 'Algorithm Quasi Newton1'
	endif

C start iterations
      do while (crit(4) .le. QN_MaxIt)
	   
	   if (QN_Global) then
           Call QNStep(nx,x1,dx,df1,f1,mstep,F,step,f2)
	     crit(3)=f2 ! save new function value
           if (step .lt. mstep) then ! return if it was not possible to decrease f along step*dx sufficiently
	       crit(1)=2
	       crit(3)=f1
	       return
	     endif
	   else
	     f2=F(nx,x1+dx)
	     step=1.0
	   endif
         x2=x1+step*dx ! next iterate of x
C	   f2=F(nx,x2)
         Call CDGrad(nx,x2,f,df2)
	   crit(2)=GradTest(nx,df2,x2,f2)
         if (QN_Print) then
	      Call SetTextPosition(2,1,curpos)
	      Write(fh,"('Iteration No. ',I5)") IDINT(crit(4))
            Call SetTextPosition(3,1,curpos)
	      Write(fh,"('Max. Rel. Grad: ',F20.15)") crit(2)
	      Call SetTextPosition(4,1,curpos)
	      Write(fh,"('Steplength:     ',F20.15)") step
	      Call SetTextPosition(5,1,curpos)
	      Write(fh,"('f(x):           ',F20.15)") crit(3)
	   Endif
	   if (crit(2) .lt. QN_GradTol) then	      
	        x1=x2
	   crit(3)=f2
	   crit(1)=0.0
	      return
         endif
C  BFGS update to inverse of approximate Hessian             
             dx=x2-x1
            ddf=df2-df1 ! the difference of the gradients
	     Call BFGS(nx,hinv,dx,ddf)
C store new x, compute new direction, store old gradient
	    x1=x2
          dx=-MatMul(hinv,df2)
	    f1=f2
	   df1=df2
       mstep=MinStep(nx,x1,dx) 
        crit(4)=crit(4)+1.0
       
	end do
      crit(1)=3
	Return

	End Subroutine



      Subroutine QuasiNewton2(f,n,x0,x,crit)
      
	use IMSLF90 
	use DFLIB	
      use Errors
	use QNP

      Integer(4) n
	Real(8) f, x0(n), x(n),crit(4)

      Logical ishere
	Integer(4) i, iercd
	INTEGER(2) fh
	Real(8) f1, f2, x1(n), h(1:n,1:n)
	Real(8) dx0(1:n,1), dx1(1:n,1)
	Real(8)  s, z, df0(1:n,1), df1(1:n,1), mstep
	Real(8) dgrad(1:n,1), hdx(1:n,1), MinStep, GradTest	
      
	External f

      Type(rccoord) curpos

C Initialize
C       Read parameters from QNP.txt (if here)
      Inquire(File='QNP.txt',Exist=ishere)
	if (ishere) then
	   Call GetD('qn_gradtol','QNP.txt',QN_GradTol)
	   Call GetD('qn_partol','QNP.txt',QN_ParTol)
	   Call GetD('qn_maxit','QNP.txt',QN_MaxIt)	   
	   Call GetL('qn_print','QNP.txt',QN_Print)
	   Call GetL('qn_global','QNP.txt',QN_Global)
      endif
C Set Default values, if not otherwise specified
      IF (QN_GradTol .lt. 0.) QN_GradTol=Epsilon(s)**(1./3.)	
	IF (QN_ParTol  .lt. 0.) QN_ParTol=Epsilon(s)**(2./3.)

      fh=GetActiveQQ() ! Handle of active window to write messages      
      Call Erset(0,0,0) ! Change error behavior of IMSL routines

C Initial function value and gradient
      FError=0
      f1=f(n,x0)

	If (FError .ne. 0) then ! return,if x0 is not admissible
	   crit(1)=1
	   return
	endif

      Call CDGrad(n,x0,f,df0)

C Test whether the current x0 is already a good solution  (as suggested by Dennis and Schnable (1983), p. 349))
      crit(2)=GradTest(n,df0,x0,f1)
	if (crit(2) .lt. (1.0e-3)*QN_GradTol) then
	   crit(1)=0.0	
	   crit(3)=f1
	   return
	endif


	H=0.0
      Do i=1,n
	   H(i,i)=1.0
	End Do

	x=x0

      if (QN_Print) then
         fh=GetActiveQQ() ! Handle of active window to write messages      
	   Call ClearScreen($GCLEARSCREEN)
	   CALL SETTEXTPOSITION(1,1,curpos)
	   WRITE(fh,*) 'Algorithm Quasi Newton2'
	endif

      crit(4)=1.0

C start iterations
      do while (crit(4) .le. QN_MaxIt)	 	

          if (QN_Print) then
            CALL SetTextPosition(2,2,curpos)
	      WRITE(fh,"('Iteration #:',I5)") IDINT(crit(4))
	    endif

          Call DLSADS(n,h,n,-df0,dx0)
	    if ((iercd() .eq. 1) .or. (iercd() .eq. 2)) then ! reinitialize h
	       h=0
	       Do i=1,n
	         h(i,i)=1.0
	       End Do
             Call DLSADS(n,h,n,-df0,dx0) ! and compute new direction
	    endif	    

	    if (QN_Global) then
	       mstep=MinStep(n,x,dx0)     
	       Call QNStep(n,x,dx0,df0,f1,mstep,F,s,f2)	       
	       if (s .lt. mstep) then
	          crit(1)=2
	          return
	       endif
	    else
	       s=1.0
	      f2=F(n,x(1:n)+dx0(1:n,1))
	    endif
	    crit(3)=f2
	    if (QN_Print) then
		   Call SetTextPosition(4,2,curpos)
	       Write(fh,"('f(x)         = ',F20.15)") crit(3)
	       Call SetTextPosition(5,2,curpos)
	       Write(fh,"('Step Length  = ',F20.15)") s	    
	    endif
	    x1=x+s*dx0(1:n,1)
          dx1(1:n,1)=x1(1:n)-x(1:n)
C	    f2=F(n,x1)
  	    crit(3)=f2
	    Call CDGrad(n,x1,f,df1)
	    dgrad=df1-df0
	    crit(2)=GradTest(n,df1,x1,f2)
	    if (QN_Print) then
	       Call SetTextPosition(6,2,curpos)
	       Write(fh,"('Max RelGrad  = ',F20.15)") crit(2)
	    endif
	    if (crit(2).lt.QN_GradTol) then
	           x=x1
           crit(1)=0.0	  	
	    return
	endif
C  BFGS update
	  hdx=MatMul(h,dx1)
	  z=Dot_product(dx1(1:n,1),hdx(1:n,1))	
	  h=h-(1/z)*MatMul(Matmul(h,dx1),matmul(transpose(dx1),h))+
     +   MatMul(dgrad,transpose(dgrad))*
     +   (1/dot_product(dgrad(1:n,1),dx1(1:n,1)))
C store new x
	     x=x1
C store new gradient
         df0=df1
C store new function value
          f1=f2
C update counter
        crit(4)=crit(4)+1.0

	End Do
	
	crit(1)=3
	
	Return
	 	
	End Subroutine


C GradP: the Gauss code to compute the gradient of a function

      Subroutine GradP(n,x0,f,df)
    
      Integer(4) n
	Real(8) x0(n), f, df(n)

      Integer(4) i
	Real(8) dh(n),ax0(n),xdh(n),arg(n),dax0(n),f0
   
      f0=f(n,x0)
	 
    
C Computation of stepsize (dh) for gradient 

      ax0 = dabs(x0)
      Do i=1,n
	  if (x0(i) .ne. 0.0) then	  
           dax0(i) = x0(i)/ax0(i)
        else
           dax0(i) = 1.0
        endif
	  dh(i)=(1.0e-8)*dmax1(ax0(i),1.0e-2)*dax0(i)
	End Do
    
      
      xdh = x0+dh
      dh = xdh-x0 ! This increases precision slightly
      
      do i=1,n
	   arg=x0
	   arg(i)=xdh(i)
	  df(i)=f(n,arg)
        df(i)=(df(i)-f0)/dh(i)
	  continue
	End Do
	
	Return
	
	End Subroutine

      Subroutine BFGS(n,h,dx,dgrad)

	Integer(4) n
	Real(8), intent(inout) :: h(n,n)
	Real(8), intent(in) :: dx(n), dgrad(n)

      Integer(4) i,j
      Logical pd
	Real(8) denom1, denom2, u(n), hdgrad(n), sumdx, sumdgrad
      Real(8) eps

      eps=1.e-8
	denom1=Dot_Product(dx,dgrad)
	hdgrad=MatMul(h,dgrad)
	denom2=Dot_Product(dgrad,hdgrad)
	sumdx=Dot_Product(dx,dx)
	sumdgrad=Dot_Product(dgrad,dgrad)
	 if (denom1 .gt. dsqrt(eps*sumdx*sumdgrad)) then
          u=(1./denom1)*dx - (1./denom2)*hdgrad
         do i=1,n
	      do j=i,n
	         h(i,j)=h(i,j)+(1./denom1)*dx(i)*dx(j)-(1./denom2)*
     +              hdgrad(i)*hdgrad(j) + denom2*u(i)*u(j)
	         h(j,i)=h(i,j)
	      End Do
	   End Do
	      pd=.true.
	      Do i=1,n ! check necessary condition for positive definite hinv
	         if (h(i,i) .lt. Epsilon(eps)) pd=.false.
	      End Do
		  if (.not.pd) then
		     h=0.0
			 Do i=1,n
			    h(i,i)=1.0
	         End Do
	      endif
	   endif

	Return

	End Subroutine